from orderly_evm_connector.lib.utils import check_required_parameters


def get_valut_balances(self, chain_id: int = None, token: str = None):
    """Get Vault Balances

    Limit: 10 requests per 1 second per IP address

    GET /v1/public/vault_balance


    Optional Args:
        chain_id(number): id of the chain you wish to query.
        token(string): the token you wish to query

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/get-vault-balance
    """
    payload = {"chain_id": chain_id, "token": token}
    return self._request("GET", "/v1/public/vault_balance", payload=payload)


# def get_valut_chain_config(self, broker_id: str = None):
#     """Get vault chain config
#
#     Limit: 10 requests per 1 second per IP address
#
#     GET /v1/public/chain_info
#
#     Args:
#         broker_id: str
#
#     https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/get-vault-chain-config
#     """
#
#     payload = {
#         "broker_id": broker_id
#     }
#     return self._request("GET", "/v1/public/chain_info", payload=payload)

def get_supported_chains_broker(self,broker_id: str = None):
    """
    Get Supported Chains per Broker
    Limit: 10 requests per 1 second per IP address

    GET /v1/public/chain_info

    Get chains specified broker is available on.
    """
    # check_required_parameters(
    #     [[broker_id, "broker_id"]]
    # )
    payload = {"broker_id":broker_id}
    return self._request("GET", "/v1/public/chain_info",payload=payload)



import asyncio
from typing import Optional

import json
import logging
from orderly_evm_connector.lib.utils import (
    get_timestamp,
    orderlyLog,
    get_uuid,
    parse_proxies,
    generate_signature,
)
from orderly_evm_connector.websocket.orderly_socket_manager import OrderlySocketManager
from orderly_evm_connector.websocket.async_websocket_manager import AsyncWebsocketManager


class OrderlyWebsocketClient:
    def __init__(
        self,
        websocket_url,
        orderly_account_id=None,
        orderly_key=None,
        orderly_secret=None,
        private=False,
        wss_id=None,
        async_mode=False,
        timeout=None,
        debug=False,
        proxies: Optional[dict] = None,
        on_message=None,
        on_open=None,
        on_close=None,
        on_error=None,
    ):
        orderly_account_id = (
            orderly_account_id
            if orderly_account_id
            else "OqdphuyCtYWxwzhxyLLjOWNdFP7sQt8RPWzmb5xY"
        )
        self.websocket_url = f"{websocket_url}/{orderly_account_id}"
        self.orderly_secret = orderly_secret
        self.wss_id = wss_id if wss_id else get_uuid()
        self.orderly_key = orderly_key
        self.private = private
        self.timeout = timeout
        self.logger = orderlyLog(debug=debug)
        self.subscriptions = []
        self._proxy_params = parse_proxies(proxies) if proxies else {}
        self.on_message = on_message
        self.on_open = on_open
        self.on_close = on_close
        self.on_error = on_error
        self.debug = debug
        if not async_mode:
            self._initialize_socket(
                self.websocket_url,
                self.wss_id,
                orderly_key,
                orderly_secret,
                on_message,
                on_open,
                on_close,
                on_error,
                timeout,
                debug,
                proxies,
            )
        self.logger.debug("Orderly WebSocket Client started.")


    async def run(self):
        manager = AsyncWebsocketManager(
            websocket_url=self.websocket_url,
            on_message=self.on_message,
            on_open=self.on_socket_open,
            on_close=self.on_close,
            on_error=self.on_error,
            debug=self.debug
        )
        asyncio.create_task(manager.run())
        await manager.ensure_init()

    def _auth_params(self):
        return {
            "id": self.wss_id,
            "event": "auth",
            "params": {
                "orderly_key": self.orderly_key,
                "sign": self._signature,
                "timestamp": self._timestamp,
            },
        }

    def _initialize_socket(
        self,
        websocket_url,
        wss_id,
        orderly_key,
        orderly_secret,
        on_message,
        on_open,
        on_close,
        on_error,
        timeout,
        debug,
        proxies,
    ):
        return OrderlySocketManager(
            websocket_url,
            on_message=on_message,
            on_open=self.on_socket_open,
            on_close=on_close,
            on_error=on_error,
            timeout=timeout,
            debug=debug,
            proxies=proxies,
        )

    def on_socket_open(self, socket_manager):
        self.logger.debug("Orderly WebSocket Connection opened. Subscribing...")
        if not hasattr(self, "socket_manager"):
            self.socket_manager = socket_manager
            self.socket_manager.start()
        if self.private:
            self.auth_login()
        for message in self.subscriptions:
            self.socket_manager.send_message(json.dumps(message))

    def auth_login(self):
        if not self.socket_manager._login:
            if self.orderly_secret:
                self._timestamp, self._signature = generate_signature(self.orderly_secret)
                self.auth_params = self._auth_params()
                self.auth_params['params']['timestamp'] = int(self.auth_params['params']['timestamp'])
            self.socket_manager.send_message(json.dumps(self.auth_params))
            self.socket_manager._login = True

    def send(self, message: dict):
        self.socket_manager.send_message(json.dumps(message))

    def send_message_to_server(self, message: dict):
        if self.private:
            self.auth_login()
        action = message["event"]
        if action and action != "unsubscribe":
            return self.subscribe(message)
        else:
            return self.unsubscribe(message)

    def subscribe(self, message):
        if str(message) not in self.subscriptions:
            self.subscriptions.append(message)
        self.socket_manager.send_message(json.dumps(message))

    def unsubscribe(self, message):
        self.socket_manager.send_message(json.dumps(message))

    def stop(self, id=None):
        self.socket_manager.close()
        self.socket_manager.join()



from orderly_evm_connector.lib.utils import check_required_parameters, get_timestamp


def get_registration_nonce(self):
    """[Public] Get Registration Nonce

    Retrieve a nonce used for registering an account on Orderly Network. The validity of the nonce value is 2 minutes. Each nonce can only be used once (ie for one account).

    GET /v1/registration_nonce

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/get-registration-nonce

    """

    return self._request("GET", "/v1/registration_nonce")

def get_account_details(self, account_id: str):
    """[Public] Get Account Details
    Limit: 10 requests per 1 second per IP address

    GET /v1/public/account

    Args：
        account_id(string):  The address of the user wallet

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/get-account-details

    """
    check_required_parameters([[account_id, "address"]])
    payload = {"account_id": account_id}
    return self._request("GET", "/v1/public/account", payload=payload)

def get_account(self, address: str, broker_id: str):
    """[Public] Get Account
    Limit: 10 requests per 1 second per IP address

    Check whether a particular account is registered on Orderly Network.

    GET /v1/get_account

    Args：
        address(string):  The address of the user wallet
        broker_id(string):	The broker that the account is registered on

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/get-account

    """
    check_required_parameters([[address, "address"], [broker_id, "broker_id"]])
    payload = {"address": address, "broker_id": broker_id}
    return self._request("GET", "/v1/get_account", payload=payload)


def get_broker(self, address: str):
    """[Public] Get Address Registered
    Limit: 10 requests per 1 second per IP address

    Check whether a particular address is registered on Orderly Network.

    GET /v1/get_broker

    Args：
        address(string):  The address of the user wallet

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/get-broker

    """
    check_required_parameters([[address, "address"]])
    payload = {"address": address}
    return self._request("GET", "/v1/get_broker", payload=payload)


def register_account(
    self, brokerId: str, chainId: int, registrationNonce: str, userAddress: str
):
    """Register Account

    Limit: 10 requests per 1 second per IP address

    Registers a new account to Orderly Network. Note an account is unique for each wallet address + broker id (ie the same wallet address can have multiple accounts with Orderly Network, 1 with each broker)

    POST /v1/register_account

    Args:
        message(json): Message object containing the message that is signed by the wallet owner
        message.brokerId(string): Broker ID
        message.chainId(number): Chain ID of registering chain (within those that are supported by the Network)
        message.timestamp(timestamp):   timestamp in UNIX milliseconds
        message.registrationNonce(string): Get Registration Nonce
        signature(string): The signature generated by signing the message object via EIP-712
        userAddress(string):   The address of the wallet signing the message object via EIP-712

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/register-account
    """
    _message = {
        "brokerId": brokerId,
        "chainId": chainId,
        "timestamp": int(get_timestamp()),
        "registrationNonce": registrationNonce,
    }
    message = {
        "domain": {
            "name": "Orderly",
            "version": "1",
            "chainId": chainId,
            "verifyingContract": "0xCcCCccccCCCCcCCCCCCcCcCccCcCCCcCcccccccC",
        },
        "message": _message,
        "primaryType": "Registration",
        "types": {
            "EIP712Domain": [
                {"name": "name", "type": "string"},
                {"name": "version", "type": "string"},
                {"name": "chainId", "type": "uint256"},
                {"name": "verifyingContract", "type": "address"},
            ],
            "Registration": [
                {"name": "brokerId", "type": "string"},
                {"name": "chainId", "type": "uint256"},
                {"name": "timestamp", "type": "uint64"},
                {"name": "registrationNonce", "type": "uint256"},
            ],
        },
    }

    _signature = self.get_wallet_signature(message=message)
    payload = {"message": _message, "signature": _signature, "userAddress": userAddress}
    check_required_parameters(
        [
            [brokerId, "brokerId"],
            [chainId, "chainId"],
            [registrationNonce, "registrationNonce"],
            [userAddress, "userAddress"],
        ]
    )
    return self._request("POST", "/v1/register_account", payload=payload)


def get_orderly_key(self, account_id: str, orderly_key: str):
    """Get Orderly Key
    Limit: 10 requests per 1 second per IP address

    GET /v1/get_orderly_key

    Check the validity of an Orderly access key attached to the account.

    Args:
        account_id(string): The account id of the user
        orderly_key(string): The public key of the Orderly access key

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/get-orderly-key
    """
    check_required_parameters(
        [[account_id, "account_id"], [orderly_key, "orderly_key"]]
    )
    payload = {"account_id": account_id, "orderly_key": orderly_key}
    return self._request("GET", "/v1/get_orderly_key", payload=payload)


def add_orderly_key(
    self,
    brokerId: str,
    chainId: int,
    orderlyKey: str,
    scope: str,
    timestamp: int,
    expiration: int,
    userAddress: str,
    **kwargs
):
    """Add Orderly Key

    Limit: 10 requests per 1 second per IP address

    POST /v1/orderly_key

    Adds an Orderly access key to an account.
    Args:
        message(json): Message object containing the message that is signed by the wallet owner
        message.brokerId(string): Broker ID
        message.chainId(number): Chain ID of registering chain (within those that are supported by the Network)
        message.orderlyKey(string):  The public key of the Orderly access key
        message.scope(string): Valid nonce from Get Registration Nonce
        message.timestamp(timestamp):   timestamp in UNIX milliseconds
        message.expiration(timestamp):	The signature generated by signing the message object via EIP-712
        signature(string): The signature generated by signing the message object via EIP-712
        userAddress(string):   The address of the wallet signing the message object via EIP-712
    Optional Args:
        tag(string): An optional tag of string values


    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/add-orderly-key
    """
    _message = {
        "brokerId": brokerId,
        "chainId": chainId,
        "orderlyKey": orderlyKey,
        "scope": scope,
        "timestamp": int(get_timestamp()),
        "expiration": expiration,
    }
    message = {
        "domain": {
            "name": "Orderly",
            "version": "1",
            "chainId": chainId,
            "verifyingContract": "0xCcCCccccCCCCcCCCCCCcCcCccCcCCCcCcccccccC",
        },
        "message": _message,
        "primaryType": "AddOrderlyKey",
        "types": {
            "EIP712Domain": [
                {"name": "name", "type": "string"},
                {"name": "version", "type": "string"},
                {"name": "chainId", "type": "uint256"},
                {"name": "verifyingContract", "type": "address"},
            ],
            "AddOrderlyKey": [
                {"name": "brokerId", "type": "string"},
                {"name": "chainId", "type": "uint256"},
                {"name": "orderlyKey", "type": "string"},
                {"name": "scope", "type": "string"},
                {"name": "timestamp", "type": "uint64"},
                {"name": "expiration", "type": "uint64"},
            ],
        },
    }
    _signature = self.get_wallet_signature(message=message)
    payload = {
        "message": _message,
        "signature": _signature,
        "userAddress": userAddress,
        **kwargs,
    }
    check_required_parameters(
        [
            [brokerId, "brokerId"],
            [chainId, "chainId"],
            [orderlyKey, "orderlyKey"],
            [scope, "scope"],
            [timestamp, "timestamp"],
            [expiration, "expiration"],
            [userAddress, "userAddress"],
        ]
    )
    return self._request("POST", "/v1/orderly_key", payload=payload)

def remove_orderly_key(self, orderly_key: str):
    """Remove Orderly Key
    Limit: 10 requests per 1 second per IP address

    POST /v1/remove_orderly_key

    Remove an Orderly access key from the account.

    Args:
        orderly_key(string): The public key of the Orderly access key

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/remove-orderly-key
    """
    check_required_parameters([[orderly_key, "orderly_key"]])
    payload = {"orderly_key": orderly_key}
    return self._sign_request("POST", "/v1/client/remove_orderly_key", payload=payload)

def update_leverage_configuration(self, leverage: int):
    """Update leverage setting

    Limit: 5 requests per 60 second per user

    POST /v1/client/leverage

    Choose maximum leverage for futures mode

    Args:
        leverage(number): One of 1,2,3,4,5,10

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/update-leverage-setting
    """
    check_required_parameters([[leverage, "leverage"]])
    payload = {"leverage": leverage}
    return self._sign_request("POST", "/v1/client/leverage", payload=payload)


def get_current_holdings(self, all: bool = None):
    """Get current holding

    Limit: 10 requests per 1 seconds

    GET /v1/client/holding

    Get the current summary of user token holdings.

    Optional Args:
        all(enum):	 true/false. If true then will return all token even if balance is empty.

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/get-current-holding
    """
    payload = {"all": all}
    return self._sign_request("GET", "/v1/client/holding", payload=payload)


def get_account_information(self):
    """Get account information

    Limit: 10 requests per 60 seconds

    Get basic account information including current user fee rates.

    GET /v1/client/info

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/get-account-information
    """
    return self._sign_request("GET", "/v1/client/info")


def set_maintenance_config(self, maintenance_cancel_order_flag: bool):
    """Set maintenance config
    Limit: 10 requests per 60 seconds

    Set the user config for whether the system should automatically cancel the user's pending orders during maintenance.

    POST /v1/client/maintenance_config

    Args:
       maintenance_cancel_order_flag : if true, system will cancel all of user's pending orders during maintenance.

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/set-maintenance-config
    """
    check_required_parameters(
        [[maintenance_cancel_order_flag, "maintenance_cancel_order_flag"]]
    )
    _payload = {"maintenance_cancel_order_flag": maintenance_cancel_order_flag}
    return self._sign_request("POST", "/v1/client/maintenance_config", payload=_payload)

def get_user_daily_statistics(self, start_date: str, end_date: str):
    """Get User Daily Statistics
    Limit 10 requests per 60 seconds

    GET /v1/client/statistics/daily

    Get user daily statistics of assets/pnl/volume.

    Args:
        start_date(string): Format YYYY-MM-DD.
        end_date(string): Format YYYY-MM-DD.
    Optional Args:
        page(number)
        size(number)

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/get-user-daily-statistics
    """
    check_required_parameters([[start_date, "start_date"], [end_date, "end_date"]])
    payload = {"start_date": start_date, "end_date": end_date}
    return self._sign_request("GET", "/v1/client/statistics/daily", payload=payload)

def get_user_daily_volume(self, start_date: str, end_date: str):
    """Get user daily volume
    Limit 10 requests per 60 seconds

    GET /v1/volume/user/daily

    Get the daily historical breakdown of the user trading volume.
    The provided start_date/end_date has to be within a 90-day range.

    Args:
        start_date(string): Format YYYY-MM-DD.
        end_date(string): Format YYYY-MM-DD.

    The provided start_date/end_date has to be within a 90-day range.

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/get-user-daily-volume
    """
    check_required_parameters([[start_date, "start_date"], [end_date, "end_date"]])
    payload = {"start_date": start_date, "end_date": end_date}
    return self._sign_request("GET", "/v1/volume/user/daily", payload=payload)


def get_user_volume_statistics(self):
    """Get user volume statistics

    Limit 10 requests per 60 seconds

    GET /v1/volume/user/stats

    Get the latest volume statistics of the user.

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/get-user-volume-statistics
    """
    return self._sign_request("GET", "/v1/volume/user/stats")


def get_current_orderlykey_info(self, **kwargs):
    """Get current Orderly key info

    Limit 10 requests per 60 seconds

    GET /v1/client/key_info

    Retrieve all the registered Orderly key pairs under the account.

    Optional Args:
        key_status(string): Filter by the status of the key (ACTIVE / REMOVING / REMOVED)

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/get-current-orderly-key-info
    """
    payload = {**kwargs}
    return self._sign_request("GET", "/v1/client/key_info", payload=payload)


def get_orderlykey_ip_restriction(self, orderly_key: str):
    """Get Orderly Key IP restriction

    Limit 10 requests per 60 seconds

    GET /v1/client/orderly_key_ip_restriction

    Retrieves the current IP restriction of a particular Orderly key.

    Args:
        orderly_key(string): The Orderly Key to query the IP restriction list

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/get-orderly-key-ip-restriction

    """
    check_required_parameters([[orderly_key, "orderly_key"]])
    payload = {"orderly_key": orderly_key}
    return self._sign_request(
        "GET", "/v1/client/orderly_key_ip_restriction", payload=payload
    )


def set_orderlykey_ip_restriction(self, orderly_key: str, ip_restriction_list: str):
    """Set Orderly key IP restriction

    Limit: 10 requests per 60 seconds

    POST /v1/client/set_orderly_key_ip_restriction

    Set the IP restrictions of a particular Orderly key under the account.

    Args:
        orderly_key(string): The Orderly Key to set the IP restriction list.
        ip_restriction(string): List of IP or IP ranges (comma separated), that will be allowed to place orders with the orderly_key.

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/set-orderly-key-ip-restriction
    """
    check_required_parameters(
        [[orderly_key, "orderly_key"], [ip_restriction_list, "ip_restriction_list"]]
    )
    payload = {"orderly_key": orderly_key, "ip_restriction_list": ip_restriction_list}
    return self._sign_request(
        "POST", "/v1/client/set_orderly_key_ip_restriction", payload=payload
    )


def reset_orderlykey_ip_restriction(self, orderly_key: str, reset_mode: str):
    """Reset orderly key IP restriction
    Limit: 10 requests per 60 seconds

    POST /v1/client/reset_orderly_key_ip_restriction

    Reset the IP restriction of a particular Orderly key under the account.

    Args:
        orderly_key(string): The Orderly Key to set the IP restriction list.
        reset_mode(string): The new mode of the IP restriction for this Orderly Key (ALLOW_ALL_IPS, DISALLOW_ALL_IPS).

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/reset-orderly-key-ip-restriction
    """
    check_required_parameters(
        [[orderly_key, "orderly_key"], [reset_mode, "reset_mode"]]
    )
    payload = {"orderly_key": orderly_key, "reset_mode": reset_mode}

    return self._sign_request(
        "POST", "/v1/client/reset_orderly_key_ip_restriction", payload=payload
    )


def get_position_history(self, symbol: str = None, limit: int  = None):
    payload = {
        "symbol": symbol,
        "limit": limit
    }

    return self._sign_request(
        "GET", "/v1/position_history", payload=payload
    )


import threading
import json
import time
from websocket import (
    create_connection,
    ABNF,
    WebSocketException,
    WebSocketConnectionClosedException,
    WebSocketTimeoutException,
)
from orderly_evm_connector.lib.utils import orderlyLog, parse_proxies, decode_ws_error_code
from orderly_evm_connector.lib.constants import (
    WEBSOCKET_TIMEOUT_IN_SECONDS,
    WEBSOCKET_FAILED_MAX_RETRIES,
    WEBSOCKET_RETRY_SLEEP_TIME,
)


class OrderlySocketManager(threading.Thread):
    def __init__(
        self,
        websocket_url,
        on_message=None,
        on_open=None,
        on_close=None,
        on_error=None,
        on_ping=None,
        on_pong=None,
        timeout=WEBSOCKET_TIMEOUT_IN_SECONDS,
        debug=False,
        proxies=None,
        max_retries=WEBSOCKET_FAILED_MAX_RETRIES,
    ):
        threading.Thread.__init__(self)
        self.websocket_url = websocket_url
        self.on_message = on_message
        self.on_open = on_open
        self.on_close = on_close
        self.on_error = on_error
        self.on_ping = on_ping
        self.on_pong = on_pong
        self.timeout = timeout
        self.logger = orderlyLog(debug=debug)
        self._proxy_params = parse_proxies(proxies) if proxies else {}
        self.subscriptions = []
        self._login = False
        self.create_ws_connection()

    def create_ws_connection(self):
        retries = 0
        while retries <= WEBSOCKET_FAILED_MAX_RETRIES:
            try:
                self.logger.debug(
                    f"Creating connection with WebSocket Server: {self.websocket_url}, proxies: {self._proxy_params}"
                )
                self.ws = create_connection(
                    self.websocket_url, timeout=self.timeout, **self._proxy_params
                )
                self.logger.debug(
                    f"WebSocket connection has been established: {self.websocket_url}, proxies: {self._proxy_params}"
                )
                self.on_open(self)
                break
            except Exception as e:
                self.logger.error(f"Failed to create WebSocket connection: {e}")
                retries += 1
                if retries <= WEBSOCKET_FAILED_MAX_RETRIES:
                    self.logger.warning(
                        f"Retrying connection... (Attempt {retries}/{WEBSOCKET_FAILED_MAX_RETRIES})"
                    )
                    time.sleep(WEBSOCKET_RETRY_SLEEP_TIME)
                else:
                    raise

    def reconnect(self):
        retries = 0
        while retries <= WEBSOCKET_FAILED_MAX_RETRIES:
            try:
                self.create_ws_connection()
                self._login = False
                break
            except Exception as e:
                self.logger.error(f"Failed to reconnect: {e}")
                retries += 1
                if retries <= WEBSOCKET_FAILED_MAX_RETRIES:
                    self.logger.warning(
                        f"Retrying connection... (Attempt {retries}/{WEBSOCKET_FAILED_MAX_RETRIES})"
                    )
                    time.sleep(WEBSOCKET_RETRY_SLEEP_TIME)
                else:
                    raise

    def send_message(self, message):
        self.logger.debug("Sending message to Orderly WebSocket Server: %s", message)
        self.ws.send(message)

    def run(self):
        self.read_data()

    def _handle_heartbeat(self):
        try:
            _payload = {"event": "pong"}
            self.ws.send(json.dumps(_payload))
            self.logger.debug(f"Sent Ping frame:{_payload}")
        except Exception as e:
            self.logger.error("Failed to send Ping: {}".format(e))

    def read_data(self):
        data = ""
        while True:
            try:
                op_code, frame = self.ws.recv_data_frame(True)
                try:
                    _message = json.loads(frame.data)
                    if "event" in _message:
                        if _message["event"] == "ping":
                            self._handle_heartbeat()
                except:
                    err_code = decode_ws_error_code(frame.data)
                    self.logger.warning(f"Websocket error code received: {err_code}")
            except WebSocketConnectionClosedException:
                self.logger.warning("WebSocket connection closed. Reconnecting...")
                self.reconnect()
                continue
            except WebSocketException as e:
                if isinstance(e, WebSocketTimeoutException):
                    self.logger.error("Websocket connection timeout")
                else:
                    self.logger.error(f"Websocket exception: {e}")
                self.reconnect()
                continue
            except Exception as e:
                self.logger.error(f"Exception in read_data: {e}")
                self.logger.warning("Reconnecting...")
                self.reconnect()
                continue
            self._handle_data(op_code, frame, data)

            if op_code == ABNF.OPCODE_CLOSE:
                self.logger.warning("CLOSE frame received, closing websocket connection")
                self._callback(self.on_close)
                break

    def _handle_data(self, op_code, frame, data):
        if op_code == ABNF.OPCODE_TEXT:
            data = frame.data.decode()
            self._callback(self.on_message, data)

    def close(self):
        if not self.ws.connected:
            self.logger.warning("Websocket already closed")
        else:
            self.ws.send_close()
        return

    def _callback(self, callback, *args):
        if callback:
            try:
                callback(self, *args)
            except Exception as e:
                self.logger.error("Error from callback {}: {}".format(callback, e))
                if self.on_error:
                    self.on_error(self, e)



from enum import Enum, auto


class AutoName(Enum):
    def _generate_next_value_(name, start, count, last_values):
        return name


class OrderType(AutoName):
    LIMIT = auto()
    MARKET = auto()
    IOC = auto()
    FOK = auto()
    POST_ONLY = auto()
    ASK = auto()
    BID = auto()


class OrderSide(AutoName):
    SELL = auto()
    BUY = auto()


class OrderStatus(AutoName):
    NEW = auto()
    CANCELLED = auto()
    PARTIAL_FILLED = auto()
    FILLED = auto()
    REJECTED = auto()
    INCOMPLETE = auto()
    COMPLETED = auto()


class TimeType(AutoName):
    _5m = auto()
    _15m = auto()
    _30m = auto()
    _1h = auto()
    _4h = auto()
    _12h = auto()
    _1d = auto()
    _1w = auto()
    _1mon = auto()
    _1y = auto()


class WalletSide(AutoName):
    DEPOSIT = auto()
    WITHDRAW = auto()

class AssetStatus(AutoName):
    NEW = auto()
    CONFIRM = auto()
    PROCESSING = auto()
    COMPLETED = auto()
    FAILED = auto()

class AlgoType(AutoName):
    STOP = auto()
    TAKE_PROFIT = auto()
    STOP_LOSS = auto()
    TP_SL = auto()
    POSITIONAL_TP_SL = auto()
    BRACKET = auto()


from orderly_evm_connector.lib.utils import check_required_parameters
from orderly_evm_connector.lib.utils import check_enum_parameter
from orderly_evm_connector.lib.enums import TimeType


def get_market_trades(self, symbol: str, limit: int = None):
    """Market trades
    Limit: 10 requests per 1 second per IP address

    GET /v1/public/market_trades

    Get the latest market trades.

    Args:
        symbol(string)
    Optional Args:
        limit(number): N (default: 10)	Numbers of trades want to query.

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/get-market-trades
    """
    check_required_parameters([[symbol, "symbol"]])
    payload = {"symbol": symbol, "limit": limit}
    return self._request("GET", "/v1/public/market_trades", payload=payload)


def get_volume_statistics(self):
    """Volume statistics

    Limit: 10 requests per 1 second per IP address

    GET /v1/public/volume/stats

    Get the latest volume statistics of Orderly and its associated brokers. Note that for broker volume, the volume is counted as the sum of the maker and taker volume, while for the full Orderly platform, the volume is the total amount matched on the platform (ie taker and maker are not double-counted.)

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/get-volume-statistics
    """
    return self._request("GET", "/v1/public/volume/stats")


def get_predicted_funding_rate_for_all_markets(self):
    """Get predicted funding rate for all markets

    Limit: 10 requests per 1 second per IP address

    GET /v1/public/funding_rates

    Get predicted funding rates for all futures trading pairs.

    Get the : * last_funding_rate : latest hourly funding rate for all the markets for the last funding period (dt = 8h) * est_funding_rate : rolling average of all funding rates over the last 8 hours

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/get-predicted-funding-rates-for-all-markets
    """
    return self._request("GET", "/v1/public/funding_rates")


def get_predicted_funding_rate_for_one_market(self, symbol: str):
    """Get predicted funding rate for one market

    Limit: 10 requests per 1 second per IP address

    GET /v1/public/funding_rate/:symbol

    Get latest funding rate for one market.

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/get-funding-rate-history-for-one-market
    """
    check_required_parameters([[symbol, "symbol"]])
    return self._request("GET", f"/v1/public/funding_rate/{symbol}")


def get_funding_rate_history_for_one_market(
    self,
    symbol: str,
    start_t: float = None,
    end_t: float = None,
    page: int = None,
    size: int = None,
):
    """Get funding rate history for one market

    Limit: 10 requests per 1 second per IP address

    GET /v1/public/funding_rate_history

    Get funding rate history for one market.

    Args:
        symbol(string)
    Optional Args:
        start_t(timestamp): start time range that you wish to query, noted that the time stamp is a 13-digits timestamp.
        end_t(timestamp): end time range that you wish to query, noted that the time stamp is a 13-digits timestamp.
        page(number): N(default: 1)	the page you wish to query.
        size(number): Default: 60

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/get-funding-fee-history
    """
    check_required_parameters([[symbol, "symbol"]])
    payload = {
        "symbol": symbol,
        "start_t": start_t,
        "end_t": end_t,
        "page": page,
        "size": size,
    }
    return self._request("GET", "/v1/public/funding_rate_history", payload=payload)


def get_futures_info_for_all_markets(self):
    """Get futures info for all markets

    Limit: 10 requests per 1 second per IP address

    GET /v1/public/futures

    Get basic futures information for all the markets.

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/get-futures-info-for-all-markets
    """
    return self._request("GET", "/v1/public/futures")


def get_futures_info_for_one_market(self, symbol: str):
    """Get futures for one market

    Limit: 10 requests per 1 second per IP address

    GET /v1/public/futures/:symbol

    Get basic futures information for one market.

    Args:
        symbol(string)

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/get-futures-info-for-one-market
    """
    check_required_parameters([[symbol, "symbol"]])
    _uri = f"/v1/public/futures/{symbol}"
    return self._request("GET", _uri)


def get_tradingview_configuration(self, locale: str):
    """Get TradingView localized configuration info

    Limit: 10 requests per 1 second per IP address

    GET /v1/tv/config

    Args:
        locale(string)

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/get-tradingview-localized-config-info
    """
    check_required_parameters([[locale, "locale"]])
    payload = {"locale": locale}
    self._request("GET", "/v1/tv/config", payload=payload)


def get_tradingview_history_basrs(
    self, symbol: str, resolution: str, from_timestamp=0, to_timestamp=0
):
    """Get TradingView history bars

    Limit: 10 requests per 1 second per IP address

    GET /v1/tv/history

    Args:
        symbol(string)
        resolution(string)
        from(timestamp)
        to(timestamp)

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/get-tradingview-history-bars

    """

    check_required_parameters(
        [
            [symbol, "symbol"],
            [resolution, "resolution"],
            [from_timestamp, "from_timestamp"],
            [to_timestamp, "to_timestamp"],
        ]
    )
    payload = {
        "symbol": symbol,
        "resolution": resolution,
        "from": from_timestamp,
        "to": to_timestamp,
    }
    self._request("GET", "/v1/tv/history", payload=payload)


def get_tradingview_symbol_info(self, group: str):
    """Get TradingView symbol info

    Limit: 10 requests per 1 second per IP address

    GET /v1/tv/symbol_info

    Args:
        group(string)

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/get-tradingview-symbol-info
    """
    check_required_parameters([[group, "group"]])
    payload = {"group": group}
    self._request("GET", "/v1/tv/symbol_info", payload=payload)


def get_orderbook_snapshot(self, symbol: str, max_level: int = None):
    """[Private] Orderbook snapshot

    Limit: 10 requests per 1 second

    GET /v1/orderbook/:symbol

    Snapshot of the current orderbook. Price of asks/bids are in descending order.

    Optional Args:
        max_level(number): (default: 100)	the levels wish to show on both side.

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/orderbook-snapshot
    """

    check_required_parameters([[symbol, "symbol"]])
    payload = {"max_level": max_level}
    return self._sign_request("GET", f"/v1/orderbook/{symbol}", payload=payload)


def get_kline(self, symbol: str, type: str, limit: int = None):
    """Get kline
    Limit: 10 requests per 1 second

    GET /v1/kline

    Get the latest klines (OHLC) of the trading pairs.

    Args:
        symbol(string)
        type(enum): 1m/5m/15m/30m/1h/4h/12h/1d/1w/1mon/1y

    Optional Args:
        limit(number): Numbers of klines. Maximum of 1000 klines.

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/get-kline
    """
    check_required_parameters([[symbol, "symbol"], [type, "type"]])
    check_enum_parameter(f"_{type}", TimeType)
    payload = {"symbol": symbol, "type": type, "limit": limit}
    return self._sign_request("GET", "/v1/kline", payload=payload)



from orderly_evm_connector.lib.utils import check_required_parameters
from orderly_evm_connector.lib.utils import check_enum_parameter
from orderly_evm_connector.lib.enums import OrderType, OrderStatus, OrderSide,AlgoType

def create_order(
    self,
    symbol: str,
    order_type: str,
    side: str,
    client_order_id: str = None,
    order_price: float = None,
    order_quantity: float = None,
    order_amount: float = None,
    reduce_only: bool = None,
    visible_quantity: float = None,
):
    """[Private] Create order

    Limit: 10 requests per 1 second

    POST /v1/order

    Args:
        symbol（string)
        order_type(enum): LIMIT/MARKET/IOC/FOK/POST_ONLY/ASK/BID
        side(enum): SELL/BUY
    Optional Args:
        client_order_id(string): 36 length, default: null
        order_price(number): If order_type is MARKET/ASK/BID, then is not required, otherwise this parameter is required.
        order_quantity(number): For MARKET/ASK/BID order, if order_amount is given, it is not required.
        order_amount(number): For MARKET/ASK/BID order, the order size in terms of quote currency
        reduce_only(boolean): Default False
        visible_quantity(number): The order quantity shown on orderbook. (default: equal to order_quantity)

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/create-order
    """
    check_required_parameters(
        [[symbol, "symbol"], [order_type, "order_type"], [side, "side"]]
    )
    check_enum_parameter(order_type, OrderType)
    payload = {
        "symbol": symbol,
        "order_type": order_type,
        "side": side,
        "client_order_id": client_order_id,
        "order_price": order_price,
        "order_quantity": order_quantity,
        "order_amount": order_amount,
        "reduce_only": reduce_only,
        "visible_quantity": visible_quantity,
    }
    return self._sign_request("POST", "/v1/order", payload=payload)

def create_algo_order(
    self,
    algo_type: str,
    quantity: int,
    side: str,
    symbol: str,
    type: str,
    trigger_price: float,
    child_orders: list = [],
    client_order_id: str = None,
    order_tag: str = None,
    price: float = None,
    reduce_only: bool = None,
    visible_quantity: bool = None,
    trigger_price_type: str = "MARK_PRICE",
):
    """[Private] Place maker/taker order that requires an additional trigger for order execution such as stop orders.

STOP type order bahavior: an order to buy or sell at the market price once the asset has traded at or through a specified price (the “stop price”). If the asset reaches the stop price, the order becomes a market order and is filled at the next available market price.

Note: This endpoint requires trading scope in Orderly Key.

    Limit: 10 requests per 1 second

    POST /v1/algo/order

    Args:
        algo_type(string)：Only STOP is available for now
        quantity(number): For MARKET/ASK/BID order, if order_amount is given, it is not required.
        side(string): SELL/BUY
        symbol(string)
        type(string):LIMIT/MARKET
        trigger_price_type(number): Only MARK_PRICE is available for now.
        trigger_price(number)

    Optional Args:
        client_order_id(string): 36 length, default: null
        order_tag(string)
        price(number)
        reduce_only(boolean): Default False
        visible_quantity(boolean): Default false

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/create-algo-order
    """
    check_required_parameters(
        [[algo_type, "algo_type"], [quantity, "quantity"], [side, "side"], [symbol, "symbol"], [type, "type"], [trigger_price, "trigger_price"]]
    )
    payload = {
        "algo_type": algo_type,
        "client_order_id": client_order_id,
        "order_tag": order_tag,
        "price": price,
        "quantity": quantity,
        "reduce_only": reduce_only,
        "side": side,
        "symbol": symbol,
        "trigger_price": trigger_price,
        "trigger_price_type": trigger_price_type,
        "type": type,
        "child_orders": child_orders,
        "visible_quantity": visible_quantity
    }
    return self._sign_request("POST", "/v1/algo/order", payload=payload)

def batch_create_order(self, orders: list):
    """[Private] Batch create order
    Limit: 1 request per 1 second

    POST /v1/batch-order

    Creates a batch of orders as a list according to the same rules as a single Create Order.

    Parameters for each order within the batch will be the same as the create single order. The batch of orders should be sent as a JSON array containing all the orders. The maximum number of orders that can be sent in 1 batch order request is 10. Each order within the batch order request is counted as 1 order towards the overall create order rate limit.

    Args:
        symbol(string)
        order_type(enum): LIMIT/MARKET/IOC/FOK/POST_ONLY/ASK/BID
        side(enum): SELL/BUY
    Optional Args:
        client_order_id(string):	36 length, default: null
        order_price	number	N	If order_type is MARKET, then is not required, otherwise this parameter is required.
        order_quantity	number	N	For MARKET/ASK/BID order, if order_amount is given, it is not required.
        order_amount	number	N	For MARKET/ASK/BID order, the order size in terms of quote currency
        reduce_only	boolean	N	Default False
        visible_quantity	number	N	The order quantity shown on orderbook. (default: equal to order_quantity)

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/batch-create-order
    """
    for order in orders:
        check_required_parameters(
            [
                [order["symbol"], "symbol"],
                [order["order_type"], "order_type"],
                [order["side"], "side"],
            ]
        )
        check_enum_parameter(order["order_type"], OrderType)

    payload = {"orders": orders}
    return self._sign_request("POST", "/v1/batch-order", payload=payload)

def edit_algo_order(
    self,
    order_id: str,
    price: float = None,
    quantity: int = None,
    trigger_price: float = None

):
    """[Private] Edit a pending algo order by order_id. Only the price or quantity can be amended.
    Limit: 10 request per 1 second

    PUT /v1/algo/order

    Edit a pending algo order by order_id. Only the price or quantity can be amended.

    Note: This endpoint requires trading scope in Orderly Key.
    Args:
        order_id(string)        
    Optional Args:
        price(number)
        quantity(number)
        trigger_price(number)
    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/edit-algo-order
    """
    check_required_parameters(
        [
            [order_id, "order_id"],
        ]
    )

    payload = {
        "order_id": order_id,
        "price": price,
        "quantity": quantity,
        "trigger_price": trigger_price
    }
    return self._sign_request("PUT", "/v1/algo/order", payload=payload)

def edit_order(
    self,
    order_id: str,
    symbol: str,
    order_type: str,
    side: str,
    client_order_id: str = None,
    order_price: float = None,
    order_quantity: float = None,
    order_amount: float = None,
    reduce_only: bool = None,
    visible_quantity: float = None,
):
    """[Private] Edit order
    Limit: 10 request per 1 second

    PUT /v1/order

    Edit a pending order by order_id. Only the order_price or order_quantity can be amended.

    Args:
        order_id(string)
        symbol(string)
        order_type(enum): LIMIT/MARKET/IOC/FOK/POST_ONLY/ASK/BID
        side(enum): SELL/BUY
    Optional Args:
        client_order_id(string):	36 length, default: null
        order_price	number	N	If order_type is MARKET, then is not required, otherwise this parameter is required.
        order_quantity	number	N	For MARKET/ASK/BID order, if order_amount is given, it is not required.
        order_amount	number	N	For MARKET/ASK/BID order, the order size in terms of quote currency
        reduce_only	boolean	N	Default False
        visible_quantity	number	N	The order quantity shown on orderbook. (default: equal to order_quantity)

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/edit-order

    """
    check_required_parameters(
        [
            [order_id, "order_id"],
            [symbol, "symbol"],
            [order_type, "order_type"],
            [side, "side"],
        ]
    )
    check_enum_parameter(order_type, OrderType)

    payload = {
        "order_id": order_id,
        "order_price": order_price,
        "order_quantity": order_quantity,
        "order_type": order_type,
        "side": side,
        "symbol": symbol,
        "client_order_id": client_order_id,
        "order_amount": order_amount,
        "reduce_only": reduce_only,
        "visible_quantity": visible_quantity,
    }
    return self._sign_request("PUT", "/v1/order", payload=payload)

def cancel_algo_order(self, order_id: int, symbol: str):
    """[Private] Cancel Algo Order

    Limit: 10 requests per 1 second

    DELETE /v1/algo/order?order_id={order_id}&symbol={symbol}

    Cancel a single algo order by order_id.

    Args:
        order_id(number)
        symbol(string)

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/cancel-algo-order
    """
    self.session.headers.update(
            {
                "Content-Type": "application/x-www-form-urlencoded"
            }
        )
    check_required_parameters([[order_id, "order_id"], [symbol, "symbol"]])
    return self._sign_request("DELETE", f"/v1/algo/order?order_id={order_id}&symbol={symbol}")

def cancel_algo_all_pending_order(self, symbol: str, algo_type: str):
    """[Private] Cancel All Pending Algo Orders

    Limit: 10 requests per 1 second

    DELETE /v1/algo/orders?symbol={symbol}

    Cancel all pending algo orders.

    Args:
        symbol(string)
        algo_type(string): STOP, TAKE_PROFIT, STOP_LOSS, TP_SL, POSITIONAL_TP_SL, BRACKET

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/cancel-all-pending-algo-orders
    """
    self.session.headers.update(
            {
                "Content-Type": "application/x-www-form-urlencoded"
            }
        )
    check_enum_parameter(algo_type, AlgoType)
    check_required_parameters([[symbol, "symbol"]])
    # add symbol and algo type if they are not None
    url = f"/v1/algo/orders"
    if symbol and algo_type:
        url += f"?symbol={symbol}&algo_type={algo_type}"
    elif symbol:
        url += f"?symbol={symbol}"
    elif algo_type:
        url += f"?algo_type={algo_type}"

    return self._sign_request("DELETE", url)

def cancel_order(self, order_id: int, symbol: str, **kwargs):
    """[Private] Cancel order

    Limit: 10 requests per 1 second

    DELETE /v1/order?order_id=:oid&symbol=:symbol

    Cancels a single order by order_id.

    Args:
        symbol(string)
        order_id(number)

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/cancel-order
    """
    check_required_parameters([[order_id, "order_id"], [symbol, "symbol"]])
    payload = {"order_id": order_id, "symbol": symbol}
    return self._sign_request("DELETE", "/v1/order", payload=payload)

def cancel_algo_order_by_client_order_id(self, client_order_id: int, symbol: str):
    """[Private] Cancel Algo Order By client_order_id

    Limit: 10 requests per 1 second

    DELETE /v1/algo/client/order?client_order_id={client_order_id}&symbol={symbol}

    Cancel an algo order by client_order_id.

    Args:
        symbol(string)
        client_order_id(string): client_order_id of the algo order

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/cancel-algo-order-by-client_order_id    """
    check_required_parameters(
        [[client_order_id, "client_order_id"], [symbol, "symbol"]]
    )
    self.session.headers.update(
            {
                "Content-Type": "application/x-www-form-urlencoded"
            }
        )
    return self._sign_request("DELETE", f"/v1/algo/client/order?client_order_id={client_order_id}&symbol={symbol}")


def cancel_order_by_client_order_id(self, client_order_id: int, symbol: str):
    """[Private] Cancel order by client_order_id

    Limit: 10 requests per 1 second

    DELETE /v1/client/order?client_order_id=:coid&symbol=:symbol

    Cancel an order by client_order_id.

    Args:
        symbol(string)
        client_order_id(string): client_order_id of the order

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/cancel-order-by-client_order_id
    """
    check_required_parameters(
        [[client_order_id, "client_order_id"], [symbol, "symbol"]]
    )
    payload = {"client_order_id": client_order_id, "symbol": symbol}
    return self._sign_request("DELETE", "/v1/client/order", payload=payload)


def cancel_orders(self, symbol: str = None):
    """[Private] Cancel orders in bulk

    Limit: 10 requests per 1 second

    DELETE /v1/orders?symbol=:symbol

    Cancel a list of orders, filtered by symbol optionally. This request will cancel all open orders within the filter criteria, and will cancel all open orders if no filter is provided.

    Optional Args:
        symbol(string)

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/cancel-orders-in-bulk
    """
    payload = {"symbol": symbol}
    return self._sign_request("DELETE", "/v1/orders", payload=payload)


def batch_cancel_orders(self, order_ids: str):
    """[Private] Batch cancel orders

    Limit: 10 requests per 1 second

    DELETE /v1/batch-order?order_ids=:order_id_1,:order_id_2

    Cancel a list of orders by order_id. The maximum orders that can be cancelled within 1 batch cancel request is 10.

    Args:
        order_ids(string): list of order_ids, comma-separated, with a maximum of 10 order ids per request.

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/batch-cancel-orders
    """
    check_required_parameters([[order_ids, "order_ids"]])
    payload = {"order_ids": order_ids}
    return self._sign_request("DELETE", "/v1/batch-order", payload=payload)


def batch_cancel_orders_by_client_order_id(self, client_order_ids: str):
    """[Private] Batch cancel orders by client_order_id

    Limit: 10 requests per 1 second

    DELETE /v1/client/batch-order?client_order_ids=:order_id_1,:order_id_2

    Cancel a list of orders by client_order_id. The maximum orders that can be cancelled within 1 batch cancel request is 10.

    Args:
        client_order_ids(string): list of order_ids, comma-separated, with a maximum of 10 order ids per request.

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/batch-cancel-orders-by-client_order_id
    """
    check_required_parameters([[client_order_ids, "client_order_ids"]])
    # payload = {"client_order_ids": ",".join(client_order_ids)}
    payload = {"client_order_ids": client_order_ids}
    return self._sign_request("DELETE", "/v1/client/batch-order", payload=payload)

def get_algo_order(self, order_id: str):
    """[Private] Get order

    Limit: 10 requests per 1 second

    GET /v1/algo/order/{order_id}

    Get details of a single algo order by order_id.

    Args:
        order_id(string): id of the algo order

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/get-algo-order
    """
    check_required_parameters([[order_id, "order_id"]])
    return self._sign_request("GET", f"/v1/algo/order/{order_id}")

def get_order(self, order_id: str):
    """[Private] Get order

    Limit: 10 requests per 1 second

    GET /v1/order/:order_id

    Get details of a single order by order_id.

    Args:
        order_id(number): id of the order

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/get-order
    """
    check_required_parameters([[order_id, "order_id"]])
    return self._sign_request("GET", f"/v1/order/{ order_id }")

def get_algo_order_by_client_order_id(self, client_order_id: str):
    """[Private] Get Algo Order by client_order_id

    Limit: 10 requests per 1 second

    GET /v1/algo/client/order/{client_order_id}

    Get details of a single algo order by client_order_id.

    Args:
        client_order_id(string): client id of the order

    """
    check_required_parameters([[client_order_id, "client_order_id"]])
    return self._sign_request("GET", f"/v1/algo/client/order/{client_order_id}")


def get_order_by_client_order_id(self, client_order_id: str):
    """[Private] Get order by client_order_id

    Limit: 10 requests per 1 second

    GET /v1/client/order/:client_order_id

    Get details of a single order by client_order_id.

    Args:
        client_order_id(string): client id of the order

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/get-order-by-client_order_id
    """
    check_required_parameters([[client_order_id, "client_order_id"]])
    return self._sign_request("GET", f"/v1/client/order/{ client_order_id }")

def get_algo_orders(
    self,
    algo_type: str,
    symbol: str = None,
    order_type: str = None,
    status: str = None,
    start_t: float = None,
    end_t: float = None,
    side: str = None,
    page: int = None,
    size: int = None,
    is_triggered: str = None,
    order_tag: str = None
    
):
    """[Private] Get Algo Orders

    Limit: 10 requests per 1 second

    GET /v1/algo/orders

    Get algo order details by customized filters.

    For filter by status, one can reference special bundled statuses below for ease of access of Open (i.e INCOMPLETE) orders or COMPLETED orders.

    INCOMPLETE = NEW + PARTIAL_FILLED

    COMPLETED = CANCELLED + FILLED

    Optional Args:
    symbol(string)
    side(string):	BUY/SELL
    order_type(string):	LIMIT/MARKET
    status(enum): NEW/CANCELLED/PARTIAL_FILLED/FILLED/REJECTED/INCOMPLETE/COMPLETED
    start_t(timestamp): start time range that wish to query, noted the time stamp is 13-digits timestamp.
    end_t(timestamp): end time range that wish to query, noted the time stamp is 13-digits timestamp.
    page(number): (default: 1)	the page you wish to query.
    size(number): (default: 25)	the page size you wish to query (max: 500)
    algo_type(string)
    is_triggered(string)
    order_tag(string)

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/get-algo-orders#openapi-evmopenapi-get-v1algoorders
    """
    if order_type:
        check_enum_parameter(order_type, OrderType)
    if side:
        check_enum_parameter(side, OrderSide)
    if status:
        check_enum_parameter(status, OrderStatus)

    check_required_parameters([[algo_type, "algo_type"]])

    payload = {
        "symbol": symbol,
        "order_type": order_type,
        "side": side,
        "status": status,
        "start_t": start_t,
        "end_t": end_t,
        "page": page,
        "size": size,
        "algo_type": algo_type,
        "is_triggered": is_triggered,
        "order_tag": order_tag
    }
    return self._sign_request("GET", "/v1/algo/orders", payload=payload)

def get_orders(
    self,
    symbol: str = None,
    order_type: str = None,
    side: str = None,
    status: str = None,
    start_t: float = None,
    end_t: float = None,
    page: int = None,
    size: int = None,
    order_tag: str = None,
    source_type: str = None,
    sort_by: str = None
):
    """[Private] Get orders

    Limit: 10 requests per 1 second

    GET /v1/orders

    Get orders by customized filters.

    For filter by status, one can reference special bundled statuses below for ease of access of Open (ie INCOMPLETE) orders or COMPLETED orders.

    INCOMPLETE = NEW + PARTIAL_FILLED
    COMPLETED = CANCELLED + FILLED

    Optional Args:
    symbol(string)
    side(string):	BUY/SELL
    order_type(string):	LIMIT/MARKET
    status(enum): NEW/CANCELLED/PARTIAL_FILLED/FILLED/REJECTED/INCOMPLETE/COMPLETED
    start_t(timestamp): start time range that wish to query, noted the time stamp is 13-digits timestamp.
    end_t(timestamp): end time range that wish to query, noted the time stamp is 13-digits timestamp.
    page(number): (default: 1)	the page you wish to query.
    size(number): (default: 25)	the page size you wish to query (max: 500)
    order_tag(string)
    sort_by(string)
    If empty:

    Descending order by updated_time if user passed update_start_t or update_end_t; descending order by order_id if updated_time are the same;

    Otherwise descending order by created_time; descending order by order_id if created_time are the same;

    If not empty:

    If sort_by == CREATED_TIME_DESC, descending order by created_time. Descending order by order_id if created_time are same;

    If sort_by == CREATED_TIME_ASC, ascending order by created_time. Ascending order by order_id if created_time are same;

    If sort_by == UPDATED_TIME_DESC, descending order by updated_time. Descending order by order_id if updated_time are same;

    If sort_by == UPDATED_TIME_ASC, ascending order by updated_time. Ascending order by order_id if updated_time are same;
    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/get-orders#openapi-evmopenapi-get-v1orders
    """
    if order_type:
        check_enum_parameter(order_type, OrderType)
    if side:
        check_enum_parameter(side, OrderSide)
    if status:
        check_enum_parameter(status, OrderStatus)

    payload = {
        "symbol": symbol,
        "order_type": order_type,
        "side": side,
        "status": status,
        "start_t": start_t,
        "end_t": end_t,
        "page": page,
        "size": size,
        "order_tag": order_tag,
        "sort_by":sort_by,
        "source_type": source_type
    }
    return self._sign_request("GET", "/v1/orders", payload=payload)


def get_all_trades_of_order(self, order_id: int):
    """[Private] Get all trades of a specific order

    Limit: 10 requests per 1 second

    GET /v1/order/:order_id/trades

    Get specific trades of an order by order_id.

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/get-all-trades-of-specific-order

    """
    check_required_parameters([[order_id, "order_id"]])
    return self._sign_request("GET", f"/v1/order/{order_id}/trades")


def get_trades(
    self,
    symbol: str = None,
    start_t: float = None,
    end_t: float = None,
    page: int = None,
    size: int = None,
):
    """[Private] Get trades

    Limit: 10 requests per 1 second

    GET /v1/trades

    Return client’s trades history within a time range.

    Optional Args:
    symbol(string)
    start_t(timestamp): start time range that wish to query, noted the time stamp is 13-digits timestamp.
    end_t(timestamp): end time range that wish to query, noted the time stamp is 13-digits timestamp.
    page(number): (default: 1)	the page you wish to query.
    size(number): (default: 25)	the page size you wish to query. (max: 500)

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/get-trades
    """
    payload = {
        "symbol": symbol,
        "start_t": start_t,
        "end_t": end_t,
        "page": page,
        "size": size,
    }
    return self._sign_request("GET", "/v1/trades", payload=payload)


def get_trade(self, trade_id: int):
    """[Private] Get trade
    Limit: 10 requests per 1 second

    GET /v1/trade/:trade_id

    Get specific transaction details by trade_id.

    Args:
        trade_id(number): id of the trade

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/get-trade
    """
    check_required_parameters([[trade_id, "trade_id"]])
    return self._sign_request("GET", f"/v1/trade/{trade_id}")


def get_all_positions_info(self):
    """Get all position info

    Limit: 30 requests per 10 second per user

    GET /v1/positions

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/get-all-positions-info
    """
    _uri = "/v1/positions"
    return self._sign_request("GET", _uri)


def get_one_position_info(self, symbol: str):
    """Get one position info

    Limit: 30 requests per 10 second per user

    GET /v1/position/:symbol

    Args:
        symbol(string)

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/get-one-position-info
    """
    check_required_parameters([[symbol, "symbol"]])
    _uri = f"/v1/position/{symbol}"
    return self._sign_request("GET", _uri)


def get_funding_fee_history(self, symbol: str, **kwargs):
    """Get funding fee history

    Limit: 20 requests per 60 second per user

    GET /v1/funding_fee/history

    Get funding fee history.

    Args:
        symbol(str)

    Optional Args:
        start_t(timestamp):	 start time range that you wish to query, noted that the time stamp is a 13-digits timestamp.
        end_t(timestamp): end time range that you wish to query, noted that the time stamp is a 13-digits timestamp.
        page(number): N(default: 1)	the page you wish to query.
        size(number): Default: 60
    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/get-funding-fee-history

    """
    check_required_parameters([[symbol, "symbol"]])
    payload = {"symbol": symbol, **kwargs}
    return self._sign_request("GET", "/v1/funding_fee/history", payload=payload)



from orderly_evm_connector.lib.utils import check_required_parameters, get_timestamp
from orderly_evm_connector.lib.utils import get_withdraw_settle_verifyingcontract


def get_settle_pnl_nonce(self):
    """[Private] Get Settle PnL nonce

    Limit: 10 requests per 1 second

    GET /v1/settle_nonce

    Retrieve a nonce used for requesting a withdrawal on Orderly Network. Each nonce can only be used once.

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/get-settle-pnl-nonce
    """
    return self._sign_request("GET", "/v1/settle_nonce")


def request_pnl_settlement(
    self,
    brokerId: str,
    chainId: int,
    settleNonce: int,
    userAddress: str,
    # verifyingContract: str,
):
    """[Private] Request PnL Settlement

    Limit: 1 requests per 1 second

    POST /v1/settle_pnl

    Args:
    message(JSON): Message object containing the message that is signed by the wallet owner
    message.brokerId(string): Broker ID
    message.chainId(int): Chain ID of registering chain (within those that are supported by the Network)
    message.settleNonce(int): once from Get settle PnL Nonce
    message.timestamp(timestamp): current timestamp in UNIX milliseconds
    signature(string): The signature generated by signing the message object via EIP-712
    userAddress(string): The address of the wallet signing the message object via EIP-712

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/request-pnl-settlement
    """

    check_required_parameters(
        [
            [brokerId, "brokerId"],
            [chainId, "chainId"],
            [settleNonce, "settleNonce"],
            [userAddress, "userAdress"],
            # [verifyingContract, "verifyingContract"],
        ]
    )
    verifyingContract = get_withdraw_settle_verifyingcontract(self.orderly_testnet)
    _message = {
        "brokerId": brokerId,
        "chainId": chainId,
        "settleNonce": settleNonce,
        "userAddress": userAddress,
        "timestamp": int(get_timestamp()),
    }
    message = {
        "domain": {
            "name": "Orderly",
            "version": "1",
            "chainId": chainId,
            "verifyingContract": verifyingContract,
        },
        "message": _message,
        "primaryType": "SettlePnl",
        "types": {
            "EIP712Domain": [
                {"name": "name", "type": "string"},
                {"name": "version", "type": "string"},
                {"name": "chainId", "type": "uint256"},
                {"name": "verifyingContract", "type": "address"},
            ],
            "SettlePnl": [
                {"name": "brokerId", "type": "string"},
                {"name": "chainId", "type": "uint256"},
                {"name": "settleNonce", "type": "uint64"},
                {"name": "timestamp", "type": "uint64"},
            ],
        },
    }
    print(message)
    _signature = self.get_wallet_signature(message=message)
    payload = {
        "message": _message,
        "signature": _signature,
        "userAddress": userAddress,
        "verifyingContract": verifyingContract,
    }
    return self._sign_request("POST", "/v1/settle_pnl", payload=payload)


def get_pnl_settlement_history(
    self, start_t: int = None, end_t: int = None, page: int = None, size: int = None
):
    """[Private] Get PnL settlement history

    Limit: 20 requests per 1 secsond

    GET /v1/pnl_settlement/history

    Retrieve the historical PnL settlement history of the account.

    Optional Args:
        start_t(timestamp)
        end_t(timestamp)
        page(int)
        size(int)
    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/get-pnl-settlement-history
    """
    payload = {"start_t": start_t, "end_t": end_t, "page": page, "size": size}
    return self._sign_request("GET", "/v1/pnl_settlement/history", payload=payload)



import json
import os
import time
import uuid

from urllib.parse import urlparse
from collections import OrderedDict
from urllib.parse import urlencode
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey
from eth_account.messages import encode_structured_data
from web3 import Web3
import base58,base64
import logging
from orderly_evm_connector.error import (
    ParameterRequiredError,
    ParameterValueError,
    ParameterTypeError,
)


import logging
initialized = False
disable_validation = os.environ.get("DISABLE_VALIDATION", False)
test_url = os.environ.get("ORDERLY_TEST_URL", False)
def orderlyLog(debug=False):
    
    logger = logging.getLogger("orderly_log")
    if not logger.handlers:
        if debug:
            logging.basicConfig(level=logging.DEBUG, format="%(asctime)s - %(levelname)s - %(message)s")
        else:
            logging.basicConfig(level=logging.WARNING, format="%(asctime)s - %(levelname)s - %(message)s")
    return logger



def cleanNoneValue(d) -> dict:
    out = {}
    for k in d.keys():
        if d[k] is not None:
            out[k] = d[k]
    return out


def check_required_parameter(value, name):
    if not value and value != 0:
        raise ParameterRequiredError([name])


def check_required_parameters(params):
    """Validate multiple parameters
    params = [
        ['PERP_BTC_USDC', 'symbol'],
        [10, 'price']
    ]
    """
    if not disable_validation:
        for p in params:
            check_required_parameter(p[0], p[1])


def check_enum_parameter(value, enum_class):
    if not disable_validation:
        if value not in set(item.value for item in enum_class):
            raise ParameterValueError([value])


def check_type_parameter(value, name, data_type):
    if value is not None and not isinstance(value, data_type):
        raise ParameterTypeError([name, data_type])


def get_timestamp():
    return int(time.time() * 1000)


def convert_list_to_json_array(symbols):
    if symbols is None:
        return symbols
    res = json.dumps(symbols)
    return res.replace(" ", "")


def get_uuid():
    return str(uuid.uuid4())


def purge_map(map: map):
    """Remove None values from map"""
    return {k: v for k, v in map.items() if v is not None and v != "" and v != 0}


def parse_proxies(proxies: dict):
    """Parse proxy url from dict, only support http and https proxy, not support socks5 proxy"""
    proxy_url = proxies.get("http") or proxies.get("https")
    if not proxy_url:
        return {}

    parsed = urlparse(proxy_url)
    return {
        "http_proxy_host": parsed.hostname,
        "http_proxy_port": parsed.port,
        "http_proxy_auth": (parsed.username, parsed.password)
        if parsed.username and parsed.password
        else None,
    }


def generate_signature(orderly_secret, message=None):
    if not orderly_secret:
        raise "Please configure orderly secret in the configuration file config.ini"
    _orderly_secret = orderly_secret.split(":")[1]
    _orderly_private_key = Ed25519PrivateKey.from_private_bytes(
        base58.b58decode(_orderly_secret)[0:32]
    )
    _timestamp = get_timestamp()
    if message and isinstance(message, dict):
        message["timestamp"] = _timestamp
    else:
        message = f"{_timestamp}{message or ''}" 
    _signature = base64.b64encode(
        _orderly_private_key.sign(bytes(message, "utf-8"))
    ).decode("utf-8")
    return str(_timestamp), _signature


def generate_wallet_signature(wallet_secret, message=None):
    private_key = f"0x{wallet_secret}"
    _message = message
    encoded_message = encode_structured_data(_message)
    w3 = Web3()
    signed_message = w3.eth.account.sign_message(
        encoded_message, private_key=private_key
    )
    return signed_message.signature.hex()


def get_endpoints(orderly_testnet):
    # True: Testnet, False: Mainnet
    if orderly_testnet and orderly_testnet.lower() == 'true':
        orderly_endpoint = test_url or "https://testnet-api-evm.orderly.org"
        orderly_websocket_public_endpoint = "wss://testnet-ws-evm.orderly.org/ws/stream"
        orderly_websocket_private_endpoint = (
            "wss://testnet-ws-private-evm.orderly.org/v2/ws/private/stream"
        )
        return (
            orderly_endpoint,
            orderly_websocket_public_endpoint,
            orderly_websocket_private_endpoint,
        )
    else:
        orderly_endpoint = "https://api-evm.orderly.org"
        orderly_websocket_public_endpoint = "wss://ws-evm.orderly.org/ws/stream"
        orderly_websocket_private_endpoint = (
            "wss://ws-private-evm.orderly.org/v2/ws/private/stream"
        )
        return (
            orderly_endpoint,
            orderly_websocket_public_endpoint,
            orderly_websocket_private_endpoint,
        )
        
def get_withdraw_settle_verifyingcontract(orderly_testnet):
    verifyingcontract  = '0x1826B75e2ef249173FC735149AE4B8e9ea10abff' if orderly_testnet else '0x6F7a338F2aA472838dEFD3283eB360d4Dff5D203'
    return verifyingcontract

def decode_ws_error_code(data):
    try:
        return str(int.from_bytes(data, byteorder='big'))
    except:
        return ""


import asyncio
from typing import Optional
from orderly_evm_connector.websocket.websocket_client import OrderlyWebsocketClient
from orderly_evm_connector.lib.utils import get_endpoints


class WebsocketPublicAPIClient(OrderlyWebsocketClient):
    def __init__(
        self,
        orderly_testnet=False,
        orderly_account_id=None,
        orderly_key=None,
        orderly_secret=None,
        private=False,
        wss_id=None,
        timeout=None,
        debug=False,
        proxies: Optional[dict] = None,
        on_message=None,
        on_open=None,
        on_close=None,
        on_error=None,
    ):
        _, self.orderly_websocket_public_endpoint, _ = get_endpoints(orderly_testnet)
        super().__init__(
            self.orderly_websocket_public_endpoint,
            wss_id=wss_id,
            private=private,
            orderly_account_id=orderly_account_id,
            on_message=on_message,
            on_open=on_open,
            on_close=on_close,
            on_error=on_error,
            timeout=timeout,
            debug=debug,
            proxies=proxies,
        )

    from orderly_evm_connector.websocket.websocket_api._stream import get_orderbook
    from orderly_evm_connector.websocket.websocket_api._stream import request_orderbook
    from orderly_evm_connector.websocket.websocket_api._stream import (
        get_orderbookupdate,
    )
    from orderly_evm_connector.websocket.websocket_api._stream import get_trade
    from orderly_evm_connector.websocket.websocket_api._stream import get_24h_ticker
    from orderly_evm_connector.websocket.websocket_api._stream import get_24h_tickers
    from orderly_evm_connector.websocket.websocket_api._stream import get_bbo
    from orderly_evm_connector.websocket.websocket_api._stream import get_bbos
    from orderly_evm_connector.websocket.websocket_api._stream import get_kline
    from orderly_evm_connector.websocket.websocket_api._stream import get_index_price
    from orderly_evm_connector.websocket.websocket_api._stream import get_index_prices
    from orderly_evm_connector.websocket.websocket_api._stream import get_mark_price
    from orderly_evm_connector.websocket.websocket_api._stream import get_mark_prices
    from orderly_evm_connector.websocket.websocket_api._stream import get_open_interest
    from orderly_evm_connector.websocket.websocket_api._stream import (
        get_estimated_funding_rate,
    )
    from orderly_evm_connector.websocket.websocket_api._stream import (
        get_liquidation_push,
    )

class WebsocketPublicAPIClientAsync(OrderlyWebsocketClient):
    def __init__(
        self,
        orderly_testnet=False,
        orderly_account_id=None,
        orderly_key=None,
        orderly_secret=None,
        private=False,
        wss_id=None,
        timeout=None,
        debug=False,
        proxies: Optional[dict] = None,
        on_message=None,
        on_open=None,
        on_close=None,
        on_error=None,
    ):
        _, self.orderly_websocket_public_endpoint, _ = get_endpoints(orderly_testnet)
        super().__init__(
            self.orderly_websocket_public_endpoint,
            wss_id=wss_id,
            private=private,
            async_mode=True,
            orderly_account_id=orderly_account_id,
            on_message=on_message,
            on_open=on_open,
            on_close=on_close,
            on_error=on_error,
            timeout=timeout,
            debug=debug,
            proxies=proxies,
        )

    # public websocket
    async def request_orderbook(self, *args, **kwargs):
        from orderly_evm_connector.websocket.websocket_api._stream import request_orderbook
        request_orderbook(self, *args, **kwargs)
        await asyncio.sleep(0)

    async def get_orderbookupdate(self, *args, **kwargs):
        from orderly_evm_connector.websocket.websocket_api._stream import (
            get_orderbookupdate,
        )
        get_orderbookupdate(self, *args, **kwargs)
        await asyncio.sleep(0)
    
    async def get_orderbook(self, *args, **kwargs):
        from orderly_evm_connector.websocket.websocket_api._stream import get_orderbook
        get_orderbook(self, *args, **kwargs)
        await asyncio.sleep(0)

    async def get_trade(self, *args, **kwargs):
        from orderly_evm_connector.websocket.websocket_api._stream import get_trade
        get_trade(self, *args, **kwargs)
        await asyncio.sleep(0)
    
    async def get_24h_ticker(self, *args, **kwargs):
        from orderly_evm_connector.websocket.websocket_api._stream import get_24h_ticker
        get_24h_ticker(self, *args, **kwargs)
        await asyncio.sleep(0)

    async def get_24h_tickers(self, *args, **kwargs):
        from orderly_evm_connector.websocket.websocket_api._stream import get_24h_tickers
        get_24h_tickers(self, *args, **kwargs)
        await asyncio.sleep(0)

    async def get_bbo(self, *args, **kwargs):
        from orderly_evm_connector.websocket.websocket_api._stream import get_bbo
        get_bbo(self, *args, **kwargs)
        await asyncio.sleep(0)

    async def get_bbos(self, *args, **kwargs):
        from orderly_evm_connector.websocket.websocket_api._stream import get_bbos
        get_bbos(self, *args, **kwargs)
        await asyncio.sleep(0)

    async def get_kline(self, *args, **kwargs):
        from orderly_evm_connector.websocket.websocket_api._stream import get_kline
        get_kline(self, *args, **kwargs)
        await asyncio.sleep(0)

    async def get_index_price(self, *args, **kwargs):
        from orderly_evm_connector.websocket.websocket_api._stream import get_index_price
        get_index_price(self, *args, **kwargs)
        await asyncio.sleep(0)

    async def get_index_prices(self, *args, **kwargs):
        from orderly_evm_connector.websocket.websocket_api._stream import get_index_prices
        get_index_prices(self, *args, **kwargs)
        await asyncio.sleep(0)

    async def get_mark_price(self, *args, **kwargs):
        from orderly_evm_connector.websocket.websocket_api._stream import get_mark_price
        get_mark_price(self, *args, **kwargs)
        await asyncio.sleep(0)

    async def get_mark_prices(self, *args, **kwargs):
        from orderly_evm_connector.websocket.websocket_api._stream import get_mark_prices
        get_mark_prices(self, *args, **kwargs)
        await asyncio.sleep(0)

    async def get_open_interest(self, *args, **kwargs):
        from orderly_evm_connector.websocket.websocket_api._stream import get_open_interest
        get_open_interest(self, *args, **kwargs)
        await asyncio.sleep(0)

    async def get_estimated_funding_rate(self, *args, **kwargs):
        from orderly_evm_connector.websocket.websocket_api._stream import (
            get_estimated_funding_rate,
        )
        get_estimated_funding_rate(self, *args, **kwargs)
        await asyncio.sleep(0)

    async def get_liquidation_push(self, *args, **kwargs):
        from orderly_evm_connector.websocket.websocket_api._stream import (
            get_liquidation_push,
        )
        get_liquidation_push(self, *args, **kwargs)
        await asyncio.sleep(0)


class WebsocketPrivateAPIClient(OrderlyWebsocketClient):
    def __init__(
        self,
        orderly_testnet=False,
        orderly_account_id=None,
        orderly_key=None,
        orderly_secret=None,
        private=True,
        wss_id=None,
        timeout=None,
        debug=False,
        proxies: Optional[dict] = None,
        on_message=None,
        on_open=None,
        on_close=None,
        on_error=None,
    ):
        _, _, self.orderly_websocket_private_endpoint = get_endpoints(orderly_testnet)
        super().__init__(
            self.orderly_websocket_private_endpoint,
            orderly_account_id=orderly_account_id,
            orderly_key=orderly_key,
            orderly_secret=orderly_secret,
            private=private,
            wss_id=wss_id,
            timeout=timeout,
            debug=debug,
            proxies=proxies,
            on_message=on_message,
            on_open=on_open,
            on_close=on_close,
            on_error=on_error,
        )

    # private websocket
    from orderly_evm_connector.websocket.websocket_api._private_stream import (
        get_account,
    )

    from orderly_evm_connector.websocket.websocket_api._private_stream import (
        get_balance,
    )
  

    from orderly_evm_connector.websocket.websocket_api._private_stream import (
        get_position,
    )
 

    from orderly_evm_connector.websocket.websocket_api._private_stream import (
        get_account_liquidations,
    )

    from orderly_evm_connector.websocket.websocket_api._private_stream import (
        get_liquidator_liquidations,
    )

    from orderly_evm_connector.websocket.websocket_api._private_stream import (
        get_wallet_transactions,
    )

    from orderly_evm_connector.websocket.websocket_api._private_stream import (
            get_pnl_settlement,
    )

    from orderly_evm_connector.websocket.websocket_api._private_stream import (
            get_notifications,
    )

    from orderly_evm_connector.websocket.websocket_api._private_stream import (
            get_execution_report,
    )

    from orderly_evm_connector.websocket.websocket_api._private_stream import (
            get_algo_execution_report,
    )

    from orderly_evm_connector.websocket.websocket_api._private_stream import (
            get_execution_report_for_single_broker,
    )

class WebsocketPrivateAPIClientAsync(OrderlyWebsocketClient):
  def __init__(
        self,
        orderly_testnet=False,
        orderly_account_id=None,
        orderly_key=None,
        orderly_secret=None,
        private=True,
        wss_id=None,
        timeout=None,
        debug=False,
        proxies: Optional[dict] = None,
        on_message=None,
        on_open=None,
        on_close=None,
        on_error=None,
    ):
        _, _, self.orderly_websocket_private_endpoint = get_endpoints(orderly_testnet)
        super().__init__(
            self.orderly_websocket_private_endpoint,
            orderly_account_id=orderly_account_id,
            orderly_key=orderly_key,
            orderly_secret=orderly_secret,
            private=private,
            wss_id=wss_id,
            timeout=timeout,
            async_mode=True,
            debug=debug,
            proxies=proxies,
            on_message=on_message,
            on_open=on_open,
            on_close=on_close,
            on_error=on_error,
        )
        
    # private websocket
  from orderly_evm_connector.websocket.websocket_api._private_stream import (
        get_account,
  )
  from orderly_evm_connector.websocket.websocket_api._private_stream import (
        get_balance,
  )
  from orderly_evm_connector.websocket.websocket_api._private_stream import (
        get_position,
  )
  from orderly_evm_connector.websocket.websocket_api._private_stream import (
        get_account_liquidations,
  )
  from orderly_evm_connector.websocket.websocket_api._private_stream import (
        get_liquidator_liquidations,
  )
  from orderly_evm_connector.websocket.websocket_api._private_stream import (
        get_wallet_transactions,
  )
  from orderly_evm_connector.websocket.websocket_api._private_stream import (
        get_pnl_settlement,
  )
  from orderly_evm_connector.websocket.websocket_api._private_stream import (
        get_notifications,
  )
  from orderly_evm_connector.websocket.websocket_api._private_stream import (
        get_execution_report,
        get_algo_execution_report,
        get_execution_report_for_single_broker
  )



from orderly_evm_connector.lib.utils import check_required_parameters, get_timestamp


def get_positions_under_liquidation(self, **kwargs):
    """Get positions under liquidation

    Limit: 10 requests per 1 second per IP address

    GET /v1/public/liquidation

    Optional Args:
        start_t(timestamp): start time range that you wish to query, noted that the time stamp is a 13-digits timestamp.
        end_t(timestamp): end time range that you wish to query, noted that the time stamp is a 13-digits timestamp.
        page(number):  (default: 1)	the page you wish to query.
        size(number):  Default: 60

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/get-positions-under-liquidation
    """
    payload = {**kwargs}
    return self._request("GET", "/v1/public/liquidation", payload=payload)


def get_liquidated_positions_info(self, symbol: str = None, start_t: int = None, end_t: int = None, page: int = None, size: int = None):
    """Get liquidated positions info

    Limit: 10 requests per 1 second per IP address

    GET /v1/public/liquidated_positions

    Args:
        symbol(string)
    Optional Args:
        tart_t(timestamp): start time range that you wish to query, noted that the time stamp is a 13-digits timestamp.
        end_t(timestamp): end time range that you wish to query, noted that the time stamp is a 13-digits timestamp.
        page(number):  (default: 1)	the page you wish to query.
        size(number):  Default: 60


    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/get-liquidated-positions-info
    """
    payload = {}
    if start_t:
        payload["start_t"] = start_t
    if end_t:
        payload["end_t"] = end_t
    if page:
        payload["page"] = page
    if size:
        payload["size"] = size
    if symbol:
        payload["symbol"] = symbol
    return self._request("GET", "/v1/public/liquidated_positions", payload=payload)


def get_insurance_fund_info(self):
    """Get insurance fund info

    Limit: 10 requests per 1 second per IP address

    GET /v1/public/insurancefund

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/get-insurance-fund-info

    """
    return self._request("GET", "/v1/public/insurancefund")


def get_liquidated_positions_by_liquidator(self, symbol: str, **kwargs):
    """Get liquidated positions by Liquidator

    Limit: 10 requests per 1 second per IP address

    GET /v1/client/liquidator_liquidations

    Args:
        symbol(string)
    Optional Args:
        start_t(timestamp): start time range that you wish to query, noted that the time stamp is a 13-digits timestamp.
        end_t(timestamp): end time range that you wish to query, noted that the time stamp is a 13-digits timestamp.
        page(number):  (default: 1)	the page you wish to query.
        size(number):  Default: 60

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/get-liquidated-positions-by-liquidator
    """
    check_required_parameters([[symbol, "symbol"]])
    payload = {"symbol": symbol, **kwargs}
    return self._sign_request(
        "GET", "/v1/client/liquidator_liquidations", payload=payload
    )


def get_liquidated_positions_of_account(self, **kwargs):
    """Get liquidated positions of account

    Limit: 10 requests per 1 second per IP address

    GET /v1/liquidations

    Optional Args:
        symbol(string)
        start_t(timestamp)
        end_t(timestamp)
        page(number)
        size(number)
    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/get-liquidated-positions-of-account
    """
    payload = {**kwargs}
    return self._sign_request("GET", "/v1/liquidations", payload=payload)


def claim_liquidated_positions(self, liquidation_id: int, ratio_qty_request, **kwargs):
    """Claim liquidated positions

    Limit: 5 requests per 1 second per IP address

    POST /v1/liquidation

    Args:
        liquidation_id(number)
        ratio_qty_request(number)
    Optional Args:
        extra_liquidation_ratio(number)
        limit_price(json): Liquidator’s instruction to let the system reject the liquidation claim if the following condition applies: if position_qty > 0, reject if mark_price > limit_price; if position_qty < 0, reject if mark_price < limit_price
        limit_price.{symbol}(number): The limit price for each symbol in the liquidation claim
        symbols: For high risk tiers only
            symbols.ratio_qty_request(number, required): Field dedicated to high risk symbols liquidations type only (symbols with liquidation tier = 2), where a liquidator can specific the ratio he wish to liquidates per symbol
            symbols.symbol(string, required): Only relevant if this is a liquidation of type high tier (symbols with liquidation tier = 2)

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/claim-liquidated-positions
    """
    check_required_parameters(
        [[liquidation_id, "liquidation_id"], [ratio_qty_request, "ratio_qty_request"]]
    )
    payload = {
        "liquidation_id": liquidation_id,
        "ratio_qty_request": ratio_qty_request,
        **kwargs,
    }
    return self._sign_request("POST", "/v1/liquidation", payload=payload)

def claim_from_insurance_fund(
    self, liquidation_id: str, symbol: str, qty_request, **kwargs
):
    """Claim from insurance fund

    Limit: 5 requests per 1 second

    POST /v1/claim_insurance_fund
    Args:
        liquidation_id(number)
        symbol(string)
        qty_request(number): Quantity to be claimed from insurance fund
    Optional Args:
        limit_price(number)

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/claim-insurance-fund
    """
    check_required_parameters(
        [
            [liquidation_id, "liquidation_id"],
            [symbol, "symbol"],
            [qty_request, "qty_request"],
        ]
    )
    payload = {
        "liquidation_id": liquidation_id,
        "symbol": symbol,
        "qty_request": qty_request,
        **kwargs,
    }
    return self._sign_request("POST", "/v1/claim_insurance_fund", payload=payload)


from orderly_evm_connector.lib.utils import check_required_parameters, get_timestamp


def get_points_epoch(self):
    """[Public] Get Number of Points for Distribution

    Retrieve the number of points for past epochs and current epoch.

    GET /v1/public/points/epoch

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/get-number-of-epoch-points

    """

    return self._request("GET", "/v1/public/points/epoch")

def get_points_epochdates(self):
    """[Public] Get Start and End Date of All Epochs

    Limit: 10 requests per 1 second per user

    GET /v1/public/points/epoch_dates

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/get-start-and-end-date-of-all-epochs#openapi-evmopenapi-get-v1publicpointsepoch_dates

    """

    return self._request("GET", "/v1/public/points/epoch_dates")

def get_user_points(self,address):
    """[Public] Get User's Points

    Limit: 10 requests per 1 second per user

    GET /v1/client/points
    Args:
        address(string)

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/get-number-of-points#openapi-evmopenapi-get-v1clientpoints
    """
    endpoint = "/v1/client/points"
    if address:
        endpoint = endpoint + f"?address={address}"
    return self._request("GET", endpoint)

def get_points_leaderboard(self,start_r: int = None,end_r: int = None,epoch_id: int = None,page: int = None,size: int = None):
    """[Public] Get Points Leaderboard
    Limit: 10 requests per 1 second per IP address  

    GET /v1/public/points/leaderboard
    Args:
        address(string)

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/get-number-of-points#openapi-evmopenapi-get-v1clientpoints
    """
    payload = {
        "start_r": start_r,
        "end_r": end_r,
        "epoch_id": epoch_id,
        "page": page,
        "size": size

    }
    return self._request("GET", "/v1/public/points/leaderboard",payload=payload)

def get_tradingrewards_epoch(self):
    """[Public] Get epochs data – return data of each epoch from epoch 1 to the current epoch
    Limit: 10 per second per IP address  

    GET /v1/public/tradingrewards/epoch_data

    """
    return self._request("GET", "/v1/public/tradingrewards/epoch_data")


def get_campaign_user_info(self, campaign_id: int, account_id: str = None, user_address: str = None, broker_id: str = None,
                           symbol: str = None, order_tag: str = None):
    """
    [Public] Get Campaign User Info
    Limit: 10 requests per second

    GET /v1/public/campaign/user

    Args:
        campaign_id: int
        account_id: str
        user_address: str
        broker_id: str
        symbol: str
        order_tag: str

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/get-campaign-user-info
    """
    payload = {
        "campaign_id": campaign_id,
        "account_id": account_id,
        "user_address": user_address,
        "broker_id": broker_id,
        "symbol": symbol,
        "order_tag": order_tag
    }
    return self._request("GET", "/v1/public/campaign/user",payload=payload)



from json import JSONDecodeError
import aiohttp
from .__version__ import __version__
from orderly_evm_connector.error import ClientError, ServerError
from orderly_evm_connector.lib.utils import (
    generate_signature,
)
from orderly_evm_connector.lib.utils import cleanNoneValue
from .api import API  # Import the original API class

class AsyncAPI(API):
    def __init__(
        self,
        orderly_key=None,
        orderly_secret=None,
        wallet_secret=None,
        orderly_testnet=False,
        orderly_account_id=None,
        proxies=None,
        timeout=None,
        debug=False
    ):
        super().__init__(
            orderly_key=orderly_key,
            orderly_secret=orderly_secret,
            wallet_secret=wallet_secret,
            orderly_testnet=orderly_testnet,
            orderly_account_id=orderly_account_id,
            proxies=proxies,
            timeout=timeout,
            debug=debug
        )
        self.headers = {
            "Content-Type": "application/json;charset=utf-8",
            "User-Agent": "orderly-connector-python/" + __version__,
        }

    async def _request(self, http_method, url_path, payload=None):
        if payload:
            _payload = cleanNoneValue(payload)
            if _payload:
                if http_method == "GET" or http_method == "DELETE":
                    url_path += "?" + "&".join(
                        [f"{k}={v}" for k, v in _payload.items()]
                    )
                    payload = ""
                else:
                    payload = _payload

        if payload is None:
            payload = ""
        url = self.orderly_endpoint + url_path
        self.logger.debug("url: " + url)
        params = cleanNoneValue(
            {
                "url": url,
                "params": payload,
                "timeout": self.timeout,
                "proxies": self.proxies,
            }
        )
        return await self._dispatch_request(http_method, params)

    async def _sign_request(self, http_method, url_path, payload=None):
        _payload = ""
        if payload:
            _payload = cleanNoneValue(payload)
            if _payload:
                if http_method == "GET" or http_method == "DELETE":
                    url_path += "?" + "&".join(
                        [f"{k}={v}" for k, v in _payload.items()]
                    )
                    _payload = ""
        params = {}
        payload = _payload if _payload else ""
        params["url_path"] = url_path
        params["payload"] = payload
        params["http_method"] = http_method
        query_string = self._prepare_params(params)
        try:
            _timestamp, _signature = generate_signature(
                self.orderly_secret, message=query_string
            )
        except ValueError:
            _timestamp, _signature = "mock_timestamp", "mock_signature"

        self.headers.update(
            {
                "orderly-timestamp": _timestamp,
                "orderly-account-id": self.orderly_account_id,
                "orderly-key": self.orderly_key,
                "orderly-signature": _signature,
            }
        )
        self.logger.debug(f"Sign Request Headers: {self.headers}")
        return await self.send_request(http_method, url_path, payload)

    async def send_request(self, http_method, url_path, payload=None):
        if payload is None:
            payload = {}
        url = self.orderly_endpoint + url_path
        self.logger.debug("url: " + url)
        params = cleanNoneValue(
            {
                "url": url,
                "params": payload,
                "timeout": self.timeout,
                "proxies": self.proxies,
            }
        )
        return await self._dispatch_request(http_method, params)

    async def _dispatch_request(self, http_method, params):
        async with aiohttp.ClientSession(headers=self.headers) as session:
            method_func = {
                "GET": session.get,
                "DELETE": session.delete,
                "PUT": session.put,
                "POST": session.post,
            }.get(http_method, session.get)

            response = None
            if http_method == "POST" or http_method == "PUT":
                response = await method_func(url=params["url"], json=params["params"])
            else:
                session.headers.update(
                    {
                        "Content-Type": "application/x-www-form-urlencoded;charset=utf-8"
                    }
                )
                response =  await method_func(url=params["url"])
            self.logger.debug("raw response from server:" + await response.text())
            await self._handle_rest_exception(response)
            try:
                data = await response.json()
            except ValueError:
                data = await response.text()
            result = data
            response.close()
            return result

    async def _handle_rest_exception(self, response):
        status_code = response.status
        if status_code <= 400:
            return
        if 400 < status_code < 500:
            try:
                err = await response.json()
            except JSONDecodeError:
                raise ClientError(
                    status_code, None, await response.text(), None, response.headers
                )
            error_data = None
            if "data" in err:
                error_data = err["data"]
            raise ClientError(
                status_code, err["code"], err["message"], response.headers, error_data
            )
        raise ServerError(status_code, await response.text())


from orderly_evm_connector.lib.utils import check_required_parameters


def get_account(self):
    """Push interval: real-time push
    https://docs-api-evm.orderly.network/#websocket-api-private-account
    """
    _message = {"id": self.wss_id, "topic": "account", "event": "subscribe"}
    self.send_message_to_server(_message)


def get_balance(self):
    """Push interval: real-time push
    https://docs-api-evm.orderly.network/#websocket-api-private-balance
    """
    _message = {"id": self.wss_id, "topic": "balance", "event": "subscribe"}
    self.send_message_to_server(_message)


def get_position(self):
    """Push interval: push on update
    https://docs-api-evm.orderly.network/#websocket-api-private-position-push
    """
    _message = {"id": self.wss_id, "topic": "position", "event": "subscribe"}
    self.send_message_to_server(_message)


def get_account_liquidations(self):
    """Push interval: push on update
    https://docs-api-evm.orderly.network/#websocket-api-private-account-liquidations
    """
    _message = {"id": self.wss_id, "topic": "liquidationsaccount", "event": "subscribe"}
    self.send_message_to_server(_message)


def get_liquidator_liquidations(self):
    """Push interval: push on addition/removal/update from list within 1s
                      1 user_id can have many liquidation_ids
    https://docs-api-evm.orderly.network/#websocket-api-private-liquidator-liquidations
    """
    _message = {
        "id": self.wss_id,
        "topic": "liquidatorliquidations",
        "event": "subscribe",
    }
    self.send_message_to_server(_message)


def get_wallet_transactions(self):
    """Push interval: real-time push on update
    https://docs-api-evm.orderly.network/#websocket-api-private-wallet-transactions
    """
    _message = {"id": self.wss_id, "topic": "wallet", "event": "subscribe"}
    self.send_message_to_server(_message)


def get_pnl_settlement(self):
    """Push interval: real-time push on update
    https://docs-api-evm.orderly.network/#websocket-api-private-pnl-settlement
    """
    _message = {"id": self.wss_id, "topic": "settle", "event": "subscribe"}
    self.send_message_to_server(_message)


def get_notifications(self):
    """Push interval: real-time push
    https://docs-api-evm.orderly.network/#websocket-api-private-notifications
    """
    _message = {"id": self.wss_id, "topic": "notifications", "event": "subscribe"}
    self.send_message_to_server(_message)

def get_execution_report(self):
    """Push interval: real-time push
    https://orderly.network/docs/build-on-evm/evm-api/introduction#websocket-api-private-notifications
    """
    _message = {"id": self.wss_id, "topic": "executionreport", "event": "subscribe"}
    self.send_message_to_server(_message)

def get_algo_execution_report(self):
    """Push interval: real-time push
    https://orderly.network/docs/build-on-evm/evm-api/introduction#websocket-api-private-notifications
    """
    _message = {"id": self.wss_id, "topic": "algoexecutionreportv2", "event": "subscribe"}
    self.send_message_to_server(_message)

def get_execution_report_for_single_broker(self, broker_id):
    """Push interval: real-time push
    https://orderly.network/docs/build-on-evm/evm-api/introduction#websocket-api-private-notifications
    """
    check_required_parameters([[broker_id, "broker_id"]])
    _message = {
        "id": self.wss_id,
        "topic": f"executionreport@{broker_id}",
        "event": "subscribe",
        "broker_id": broker_id,
    }
    self.send_message_to_server(_message)